import sys
import sysconfig

from path import Path

import alpakka.commands
import alpakka.pyang_plugins
import alpakka.wools


# auto-load all Wools which are defined as 'alpakka_wools' entry points in
# their distribution's setuptools.setup()
alpakka.wools.load_from_entry_points()


# Parse command line for pure non-pyang-invoking alpakka command flags
ALPAKKA_COMMAND_ARGS, _ = alpakka.commands.PARSER.parse_known_args()

ALPAKKA_COMMAND = ALPAKKA_COMMAND_ARGS.command
if ALPAKKA_COMMAND:
    sys.exit(ALPAKKA_COMMAND())


# If no pure alpakka command flag was given then invoke pyang...
ALPAKKA_PLUGIN_DIR = Path(alpakka.pyang_plugins.__file__).realpath().dirname()

PYANG_SCRIPT = Path(sysconfig.get_path('scripts')) / 'pyang'


sys.argv += ['--plugindir', ALPAKKA_PLUGIN_DIR, '-f', 'alpakka']
exec(compile(PYANG_SCRIPT.text(), PYANG_SCRIPT, 'exec'))



import sys
from collections import OrderedDict

import alpakka
from alpakka.wools import Wool

WOOLS = alpakka.WOOLS


class NodeWrapperMeta(type):
    """
    Metaclass for :class:`NodeWrapper`

    Allows to add the YANG node name to the definition of a derived class,
    which automagically registers that class in its Wool.

    See :class:`NodeWrapper` for usage examples
    """

    mixins = {}

    def __new__(mcs, clsname, bases, clsattrs, *args, yang=None):
        # To support @NodeWrapper.mixin classes in bases, we need a new
        # metaclass additionally derived from all the metaclasses of the
        # bases, but w/o duplicates or metaclasses that are superclasses of
        # others in the list of base metaclasses
        metabases = [mcs]
        for meta in map(type, bases):
            metabases = [mb for mb in metabases if not issubclass(meta, mb)]
            if meta not in metabases:
                metabases.append(meta)

        class Meta(*metabases):
            mixins = {}

        return type.__new__(Meta, clsname, bases, clsattrs, *args)

    def __init__(cls, clsname, bases, clsattrs, *args, yang=None):
        type.__init__(cls, clsname, bases, clsattrs, *args)

        module = sys.modules.get(cls.__module__)
        if module is None:
            return

        # To get registered, the class must NOT be defined in sub-packages
        # (sub-directories) of this basic wrapper package or a wool package.
        # Only classes from sub-modules (.py files) are allowed
        pkgname = module.__package__
        if pkgname == sys.modules[__name__].__package__:
            wool = getattr(WOOLS, 'default', None)
            if wool is None:
                wool = WOOLS.default = Wool('default', __package__)
        elif pkgname in WOOLS:
            wool = WOOLS[pkgname]
        else:
            return

        # Give every class reference to its Wools for easy access
        cls.WOOL = wool

        wrapdict = wool.yang_wrappers

        # If we have an explicit class SomeWrapper(..., yang=<yang name>)
        # relation, then just add it to the wrapper dict
        if yang:
            wrapdict[yang] = cls
            return

        # Otherwise look if any base class already exists in the wrapper dict
        # and exchange it accordingly
        wrapitems = list(wrapdict.items())
        for base in cls.mro()[1:]:
            for yangname, wrapcls in wrapitems:
                if wrapcls is base:
                    wrapdict[yangname] = cls

    def mixin(cls, mixincls):
        """
        Class decorator for defining YANG node wrapper plugins in Wools.

        Any class in the wrapper hierarchy can be extended with mixins:

        >>> @NodeWrapper.mixin
        ... class SomeMixin:
        ...     def upper_name(self):
        ...         return self.yang_name().upper()

        Mixins are stored by their Wool package name in the wrapper class'
        meta ``.mixins`` dictionary:

        >>> NodeWrapper.mixins
        {...'alpakka.wrapper': [<class '...SomeMixin'>]...}

        Accessing a wrapper class through a Wool will automagically create a
        derived class with all the applicable mixins as additional (higher
        priority) base classes

        This dynamic class creation process is cached. If there might be a
        need to clear the cache for whatever reason, you can do so with:

        >>> Wool._woolify.cache_clear()

        The new class gets a special ``.__module__`` name referring to the
        alpakka.WOOLS registry:

        >>> WOOLS.default.Module.mro()
        [<class '...WOOLS['default'].Module'>, <class '...SomeMixin'>, ...]

        And after instantiation, all additional mixin features are available:

        >>> _dummy = __import__('pyang').statements.Statement(
        ...     None, None, None, 'module', 'some-module')

        >>> wrapped_module =  WOOLS.default.Module(_dummy)
        >>> wrapped_module.yang_name()
        'some-module'
        >>> wrapped_module.upper_name()
        'SOME-MODULE'

        :return: The original `mixincls` object
        """
        pkgname = sys.modules[mixincls.__module__].__package__
        if pkgname == sys.modules[__name__].__package__ or pkgname in WOOLS:
            type(cls).mixins.setdefault(pkgname, []).append(mixincls)
        return mixincls


class NodeWrapper(metaclass=NodeWrapperMeta):
    """
    Base class for node wrappers. It includes the base setup.

    If a derived classes directly maps to a YANG node, its name can be given
    as additional keyword-only argument to the class definition:

    >>> class SomeNode(NodeWrapper, yang='some-node'):
    ...     pass

    This will automagically register that class its Wool. In this case to
    ``WOOLS.default``, since we are in the default wool package and not in a
    customized wool:

    >>> WOOLS.default['some-node']
    <class 'alpakka.WOOLS['default'].SomeNode'>
    >>> WOOLS.default['some-node'].mro()
    [..., <class 'alpakka.wrapper.nodewrapper.SomeNode'>, ...]
    """
    prefix = ""

    def __init__(self, statement, parent=None):
        """
        The Constructor for NodeWrapper object.

        Attributes which are common for all wrapped statements
        :param statement: the original statement generated from pyang
        :param parent: the wrapped parent statement of the current statement,
                None for the top element
        :param yang_stmt_type: yang type of the current statment, like module,
                container, list, leaf and so on
        :param yang_stmt_name: name of the current statement, is the same
                like in the original yang file
        :param description: string containing the description of the current
                statement provided by the yang file
        :param config: if the current statement has config substatement the
                value this one is stored in this variable
        """

        self.statement = statement
        self.parent = parent
        self.is_augmented = False
        for stmt in statement.substmts:
            if stmt.keyword == 'description' and stmt.arg.lower() != "none":
                self.description = stmt.arg
            elif stmt.keyword == 'config':
                self.config = stmt.arg.lower() == 'true'
        if self.top() is not self and self.yang_type() not in ('enum', 'input',
                                                               'output',
                                                               'type'):
            nodes = self.top().all_nodes.setdefault(statement.keyword,
                                                    OrderedDict())
            nodes[self.generate_key()] = self

    def yang_name(self):
        return self.statement.arg

    def yang_type(self):
        return self.statement.keyword

    def yang_module(self):
        return (self.statement.top or self.statement).i_modulename

    def top(self):
        """
        Find the root wrapper object by walking the tree recursively
        to the top.
        :return: the root node
        """
        if self.parent:
            return self.parent.top()
        else:
            return self

    def generate_key(self):
        """
        Generates a string which represents the path from the root element of
        the Statement to the target as uniquekey. In the case of imported
        Statements the string represents only the top element and the target
        element name

        :return: a unique key which is a human readable path to the module
        """
        key = ''
        # Key generation for elements which could be imported from other
        # Modules or be implemented locally
        if self.yang_type() == 'grouping' or self.yang_type() == 'typedef':
            return self.statement.parent.arg + "/" + self.statement.arg
        # Key generation for all other Statements
        else:
            if self.parent:
                key = self.parent.generate_key() + '/'

            return key + self.yang_name()


class Listonder():
    """
    Metaclass for all List type objects
    """

    def min_list_elements(self):
        """
        Method to return the minimum occurrence of list objects, if the
        respective substmt is present

        return: the number of minimum occurrence of list objects
        """
        for i in self.statement.search('min-elements'):
            return i.arg

    def max_list_elements(self):
        """
        Method to return the maximum occurrence of list objects, if the
        respective substmt is present

        return: the number of maximum occurrence of list objects
        """
        for i in self.statement.search('max-elements'):
            return i.arg



import logging

formatter = logging.Formatter("alpakka [%(module)s, %(levelname)s]: "
                              "%(message)s")
handler = logging.StreamHandler()
handler.setFormatter(formatter)

LOGGER = logging.getLogger('alpakka')
LOGGER.setLevel(logging.INFO)
LOGGER.addHandler(handler)



from pkg_resources import iter_entry_points
from pprint import pformat

from .default_wool import Wool
from alpakka.logger import LOGGER

__all__ = ['Wool', 'WoolsRegistry']


def load_from_entry_points():
    """
    Auto-load all Wool packages which are defined as ``'alpakka_wools'``
    entry-points in their distribution's ``setuptools.setup()``, which looks
    like::

       setup(
           entry_points={
               'alpakka_wools': ['name=package'],
               ...},
           ...)

    To be actually registered as Wools, those packages must of course
    additionally call :func:`alpakka.register_wool`
    """
    for woolpoint in iter_entry_points('alpakka_wools'):
        woolpoint.load()


class WoolsRegistry:
    """
    The class of the :data:`alpakka.WOOLS` registry

    Actual wool registration is done via :func:`alpakka.register_wool`. See
    the latter for more information about that process. It also explains how
    to access the registered :class:`alpakka.Wool` instances later

    The manager registers every wool with the normalized all-lowercased wool
    name (wool id) as plugin name
    """

    def __init__(self):
        self.wools = dict()

    def register(self, wool):
        """
        Registers the wool together with its python package

        :param wool: the wool
        """
        self.wools[wool.package] = wool
        LOGGER.info("Registered {!r}".format(wool))

    def items(self):
        """
        Iterates ``(<wool id>, <alpakka.Wool instance>)`` pairs
        """
        for wool in self.wools.items():
            yield wool.id(), wool

    def names(self):
        """
        Iterates names of registered wools
        """
        for wool in self.wools.items():
            yield wool.name

    def packages(self):
        """
        Iterates python package names of registered wools
        """
        yield from self.wools.keys()

    def __contains__(self, name_or_package):
        """
        Checks if given string is either a name or a python package name of
        any registered wool, whereby name lookup is case-insensitive
        """
        for package, wool in self.wools.items():
            if name_or_package == package or (
                    name_or_package.lower() == wool.id()):
                return True

        return False

    def __getitem__(self, name_or_package):
        """
        Get a registered :class:`alpakka.Wool` instance by either its name or
        its python package name, whereby name lookup is case-insensitive
        """
        wool = self.wools.get(name_or_package)
        if wool:
            return wool

        for i_wool in self.wools.values():
            if i_wool.id() == name_or_package.lower():
                return i_wool

        raise KeyError(name_or_package)

    def __repr__(self):
        """
        Pretty-print a ``dict``-representation of the registered wools with
        their normalized identification names as keys
        """
        return pformat(dict(self.items()))



from alpakka.wrapper import wrap_module
from alpakka.wrapper.grouponder import Module
from collections import OrderedDict


def test_wrap_module(yang_module):
    """
    Test function to test the correct wrapping of a module statement

    :param yang_module:         python representation of a yang module
                                generated with pyang (see conftest.py)
    :return:                    no return value implemented
    """
    wrapped_module = wrap_module(yang_module)
    assert isinstance(wrapped_module, Module)

    # check the mandatory fields of an module
    assert wrapped_module.yang_name() == yang_module.arg
    assert wrapped_module.yang_type() == yang_module.keyword
    assert (wrapped_module.description ==
            yang_module.search_one('description').arg)


class TestModule:

    def test_all_nodes(self, yang_module):
        wrapped_module = wrap_module(yang_module)
        assert 'module' not in wrapped_module.all_nodes
        assert {'topologies', 'endpoints'} == {
            i.rsplit('/', 1)[1] for i in
            wrapped_module.all_nodes['container'].keys()}
        assert {'intent-group'} == {
            i.rsplit('/', 1)[1] for i in
            wrapped_module.all_nodes['grouping'].keys()}
        assert {'endpoints'} == {
            i.rsplit('/', 1)[1] for i in
            wrapped_module.all_nodes['leaf-list'].keys()}
        assert {
            'intents', 'installed-topologies', 'assigned-endpoints',
            'intents'} == {
            i.rsplit('/', 1)[1] for i in
            wrapped_module.all_nodes['list'].keys()}
        assert {
            'endpoint-identifier', 'intent-identifier',
            'topology-identifier', 'bit-rate', 'disjoint'} == {
            i.rsplit('/', 1)[1] for i in
            wrapped_module.all_nodes['typedef'].keys()}
        assert {
            'endpoint-id', 'protection', 'minimum-paths',
            'flexible-bandwidth', 'dedicated-bandwidth',
            'intent-id', 'disjoint-paths', 'satisfied',
            'maximum-active-connections',
            'topology-id'} == {
            i.rsplit('/', 1)[1] for i in
            wrapped_module.all_nodes['leaf'].keys()}

    def test_yang_type(self, yang_module):
        wrapped_module = wrap_module(yang_module)
        assert 'module' == wrapped_module.yang_type()

    def test_yang_name(self, yang_module):
        wrapped_module = wrap_module(yang_module)
        assert 'virtual-topology-api' == wrapped_module.yang_name()

    def test_statement(self, yang_module):
        wrapped_module = wrap_module(yang_module)
        assert yang_module == wrapped_module.statement

    def test_parent(self, yang_module):
        wrapped_module = wrap_module(yang_module)
        assert None is wrapped_module.parent

    def test_description(self, yang_module):
        wrapped_module = wrap_module(yang_module)
        assert yang_module.search_one('description').arg == \
            wrapped_module.description

    def test_children(self, yang_module):
        wrapped_module = wrap_module(yang_module)
        assert set([i.arg for i in yang_module.i_children]) == set(
            wrapped_module.children.keys())

    def test_derived_types(self, yang_module):
        wrapped_module = wrap_module(yang_module)
        assert yang_module.i_typedefs.keys() == \
            wrapped_module.derived_types.keys()

    def test_uses(self, yang_module):
        wrapped_module = wrap_module(yang_module)
        assert OrderedDict() == \
            wrapped_module.uses

    def test_all_children(self, yang_module):
        wrapped_module = wrap_module(yang_module)
        assert wrapped_module.children == wrapped_module.all_children()

    def test_top(self, yang_module):
        wrapped_module = wrap_module(yang_module)
        assert wrapped_module == wrapped_module.top()


class TestContainer:

    def test_parent(self, yang_module):
        """
        Test function to test the correct wrapping of all container statements
        attached to the yang module under test

        :param yang_module:         python representation of a yang module
                                    generated with pyang (see conftest.py)
        :return:
        """
        wrapped_module = wrap_module(yang_module)
        wrapped_container = wrapped_module.topologies

        assert wrapped_module == wrapped_container.parent

    def test_children(self, yang_module):
        wrapped_module = wrap_module(yang_module)
        wrapped_container = wrapped_module.topologies

        assert sorted(['installed-topologies']) == sorted(wrapped_container.
                                                          children)

    def test_config(self, yang_module):
        wrapped_module = wrap_module(yang_module)
        wrapped_container = wrapped_module.topologies

        assert True is wrapped_container.config



__all__ = ('NodeWrapper', 'wrap_module', 'Grouponder', 'Typonder')

import alpakka

from .nodewrapper import NodeWrapper
from .grouponder import Grouponder
from .typonder import Typonder


def wrap_module(statement, wool=None):
    """
    Wrap the module statement for akka code generation.

    :param statement: the module statement
    :return:          the wrapped module representation
    """
    if wool is None:
        wool = alpakka.WOOLS.default
    return wool['module'](statement)



from alpakka.logger import LOGGER
import optparse
from pyang import plugin

from IPython import start_ipython

import alpakka
from alpakka import WOOLS
from alpakka.wrapper import wrap_module

default_values = {
    'int': 0,
    'boolean': 'false'
}


def pyang_plugin_init():
    """
    Called by pyang plugin framework at to initialize the plugin.
    :return:
    """
    plugin.register_plugin(AlpakkaPlugin())


class AlpakkaPlugin(plugin.PyangPlugin):
    """
    Plugin to convert a yang model to code stubs.
    """

    def __init__(self):
        super().__init__()

    def add_output_format(self, fmts):
        self.multiple_modules = True
        fmts['alpakka'] = self

    def add_opts(self, optparser):
        """
        Add akka specific options to the pyang CLI.
        :param optparser: the option parser
        """
        options = [
            optparse.make_option(
                "-w", "--wool", action="store", dest="wool",
                help="wool used for knitting the code. Specifies the wool "
                     "required for the target programming language and "
                     "framework."
            ),
            optparse.make_option(
                "--output-path", dest="output", action="store",
                help="specifies the root directory for the code generation"
            ),
            optparse.make_option(
                "-i", "--interactive", action="store_true", dest="interactive",
                default=False,
                help="run alpakka in interactive mode by starting an IPython "
                     "shell before template generation"
            ),
            optparse.make_option(
                "--configuration-file-location", action="store",
                dest="config_file", help="path of the wool configuration file"
            )
        ]
        group = optparser.add_option_group("Akka output specific options")
        group.add_options(options)

    def emit(self, ctx, modules, writef):
        """
        Method which is called by pyang after the parsing and validation is
        finished, this method initiates the output conversion
        :param ctx:
        :param modules: Array of statement objects representing all modules
        :param writef:
        """
        self.get_options(ctx)
        unique_modules = set()
        # collect set of unique modules, avoid wrapping the same one
        # multiple times
        for module in modules:
            for module_details, context_module in module.i_ctx.modules.items():
                unique_modules.add(context_module)
        # wrap unique modules
        wrapped_modules = dict()
        for module in unique_modules:
            LOGGER.info("Wrapping module %s (%s)",
                        module.arg, module.i_latest_revision)
            # wrap module statement
            wrapped_module = wrap_module(module, wool=self.wool)
            wrapped_modules[wrapped_module.yang_module()] = wrapped_module

        # due to the wrapping process statements which are used multiple times
        # in different modules might be wrapped multiple times. To avoid
        # multiple output generations the data structure is traversed and all
        # duplicates, in modules which are not the module which was originally
        # implementing the statement, are deleted.
        # Is implemented inside the used wool and can be wool specific
        for module in wrapped_modules.values():
            self.wool.wrapping_postprocessing(module, wrapped_modules)

        if ctx.opts.interactive:
            start_ipython([], user_ns=dict(
                ((module.statement.arg.replace('-', '_'), module)
                 for module in wrapped_modules.values()),
                alpakka=alpakka,
            ))
        else:
            # output generation is performed per parsed module and is
            # implemented inside the used wool. Can be wool specific
            for wrapped_module in wrapped_modules.values():
                self.wool.generate_output(wrapped_module)

    def get_options(self, ctx):
        """
        Extract option parameters from the context.
        :param ctx: the context
        """
        # parsing of options which are general for the alpakkaplugin and not
        # wool specific
        self.wool = WOOLS[ctx.opts.wool] or WOOLS.default
        # set output path
        self.wool.output_path = ctx.opts.output or ""
        # pasing from the wool specific options and configuration parameters
        # is implemented inside the specific wool and can be wool specific
        self.wool.parse_config(ctx.opts.config_file)



from alpakka.logger import LOGGER
from functools import lru_cache
import re
import pyang.types


class Types:
    """
    Class storing mappings from yang data types to language specific data
    types.
    """

    def __init__(self, patterns):
        """
        Initialize types with a set of tuples, which comprise a regular
        expression and a corresponding type.

        :param patterns: A set of tuples storing patterns and types
        """
        self.patterns = patterns

    def __getitem__(self, yang_type):
        for yang_pattern, data_type in self.patterns:
            if re.match(yang_pattern, yang_type):
                return data_type
        if pyang.types.is_base_type(yang_type):
            return yang_type


class Wool(object):
    """
    The wool handler that stores the wool's name, its python package name,
    and an optional parent :class:`Wool` instance

    For case-insentitive lookups, a normalized identification name for the
    wool is created by lowering the wool's name
    """

    def __init__(self, name, package, parent=None, type_patterns=None):
        """
        :param name:    The wool's name
        :param package: The wool's fully-qualified python package name
        :param parent:  The optional :class:`Wool` instance of a parent wool
                        which is further customized by this wool
        """
        self.name = name
        self.package = package
        self.parent = parent
        self.output_path = ''
        self.config = {}
        self.data_type_mappings = Types(type_patterns or {})
        self.yang_wrappers = parent and dict(parent.yang_wrappers) or {}
        LOGGER.debug("Wool created: %s" % name)

    def id(self):
        """
        :return: the wool's identifier (lower case name)
        """
        return self.name.lower()

    @lru_cache()
    def _woolify(self, wrapcls):
        """
        Woolify the given :class:`alpakka.NodeWrapper`-derived `wrapcls`.

        Dynamically derive a new class from `wrapcls` and all ``.mixins``
        for this Wool and its parents defined by `wrapcls` and all its bases

        Finally add a ``.WOOL`` reference to the new class

        :return: A cached woolified class derived from `wrapcls`
        """
        # Can't be imported at module level due to unsatisfied circular
        # dependencies
        import alpakka.wrapper

        mixins = []
        for base in wrapcls.mro():
            if issubclass(base, alpakka.wrapper.NodeWrapper):
                wool = self
                while wool is not None:
                    for mixincls in base.mixins.get(wool.package, ()):
                        mixins.append(mixincls)
                    wool = wool.parent

        bases = (*mixins, wrapcls)
        # We also need a new metaclass derived from all the metaclasses of the
        # bases, but w/o duplicates or metaclasses that are superclasses of
        # others in the list of base metaclasses
        metabases = []
        for meta in map(type, bases):
            metabases = [
                mb for mb in metabases
                if not issubclass(meta, mb)]
            if meta not in metabases:
                metabases.append(meta)

        class Meta(*metabases):
            pass

        return Meta(wrapcls.__name__, (*mixins, wrapcls), {
            '__module__': "alpakka.WOOLS['{}']".format(self.name),
            'WOOL': self})

    def __getitem__(self, name):
        """
        Get the Woolified YANG node wrapper class for YANG statement `name`.

        See :meth:`._woolify` for Woolification details
        """
        return self._woolify(self.yang_wrappers[name])

    def get(self, name):
        """
        Get the Woolified YANG node wrapper class for YANG statement `name`.

        Returns ``None`` if no wrapper can be found instead of raising an
        exception

        See :meth:`._woolify` for Woolification details
        """
        wrapcls = self.yang_wrappers.get(name)
        if wrapcls is not None:
            return self._woolify(wrapcls)

    def __getattr__(self, name):
        """
        Get the Woolified YANG node wrapper class with given `name`.

        See :meth:`._woolify` for Woolification details
        """
        for wrapcls in self.yang_wrappers.values():
            if wrapcls.__name__ == name:
                return self._woolify(wrapcls)

        raise AttributeError(name)

    def __dir__(self):
        """
        Add all YANG node wrapper class names of this Wool.

        For use with :meth:`.__getattr__`
        """
        return (*super().__dir__(),
                *(wrapcls.__name__ for wrapcls in
                  getattr(self, 'yang_wrappers', dict()).values()))

    def __repr__(self):
        """
        Create a representation in instantiation code style
        """
        return "{}({!r}, {!r}, parent={!r})".format(
            type(self).__qualname__, self.name, self.package,
            self.parent and self.parent.name)

    def parse_config(self, path):
        """
        Method interface for the wool implementation to handle the wool
        specific configuration and options
        :param path: location of the configuration file
        :return
        """
        raise NotImplementedError

    def generate_output(self, module):
        """
        Method interface for the wool implementation of the output generation
        is implemented as part of the dedicated wool implementation

        :param module: wrapped module for which the output is generated
        :return
        """
        raise NotImplementedError

    def wrapping_postprocessing(self, module, wrapped_modules):
        """
        Method Interface for the duplication detection and the data cleansing
        which can be used and implemented by each wool

        :param wrapped_modules: list of all module statement which are
            processed
        :param module: the module for which the cleansing should be performed
        :return
        """
        raise NotImplementedError



__all__ = ['WOOLS', 'LOGGER', 'NodeWrapper', 'Wool', 'run']

from alpakka.logger import LOGGER
from alpakka.wools import WoolsRegistry, Wool

#: The central Wool registry. ``WOOLS.default`` is the basic Wool. All Wool
#  plugins go into ``WOOLS[<plugin name>]``
WOOLS = WoolsRegistry()

# implicitly loads default Wool
from alpakka.wrapper import NodeWrapper  # Ignore PycodestyleBear (E402)


def run():
    from . import __main__  # which exec()s the pyang script

    # then manually run the pyang script's run function
    # which is only run automatically if __file__ == '__main__'
    # but now __main__.__file__ == 'alpakka.__main__'
    __main__.run()



from argparse import ArgumentParser

import alpakka


PARSER = ArgumentParser('alpakka')


def command(func):
    """
    Decorator for registering `func` as alpakka command, which be invoked from
    the command line like::

       alpakka --func-name

    Whereby underscores in the function name are turned into hyphens

    Every command function must return an alpakka exit code number
    """
    PARSER.add_argument(
        '--' + func.__name__.replace('_', '-'), dest='command',
        action='store_const', const=func, default=None,
        help=func.__doc__)
    return func


@command
def list_wools():
    """
    List all registered Wools for knitting code
    FIXME: Does not work, missing ID
    """
    for name, wool in alpakka.WOOLS.items():
        print("[{}] {!r}".format(name, wool))
    return 0



import pytest
import pyang
from path import Path

"""
    Pytest fixture statement, works like a global Variable for the test
    Session. Provides a fixed baseline upon which tests can reliably and
    repeatedly be executed.
    In this file should only statements be defined which are used in multiple
    tests and for the testing of different modules.
"""


@pytest.fixture(scope='session')
def yang_module_name():
    """
    Fixture statement to define the name of reference yang input file and the
    expected yang module which will be processed for testing purposes

    :return: name of the yang module
    """
    return 'virtual-topology-api'


@pytest.fixture(scope='session')
def pyang_plugin_name():
    """
    Fixture statement which returns the name of the pyang plugin which will
    be used
    during the test session

    :return: name of the pyang plugin
    """
    return 'akka'


@pytest.fixture(scope='session')
def yang_module(yang_module_name, yang_context):
    """
    Fixture statement to generate a pyang representation of a yang_module which
    should be used for testing.
    The functionality of the statement is the following, use the pyang
    context generated at the yang_context() function. Afterwards the module
    under test is added to the context.

    :param yang_module_name:    name of the module under test
    :param yang_context:        test context
    :return:                    the module instance
    """

    path = Path(__file__).dirname() / yang_module_name + '.yang'

    module = yang_context.add_module(path, path.text())

    return module


@pytest.fixture(scope='session')
def yang_context():
    """
    Fixture statement to generate the pyang context for test purposes. A
    context is a module which encapsulates the parsing session.

    :return: pyang context object
    """
    repo = pyang.FileRepository(".")

    ctx = pyang.Context(repo)

    return ctx


@pytest.fixture(scope='session')
def yang_query():
    def query(statement, yang_type):
        node_list = set(getattr(statement, 'i_children', ()))
        node_list.intersection_update(getattr(statement, 'substmts', ()))
        for stmt in node_list:
            if stmt.keyword == yang_type:
                yield stmt
            yield from query(stmt, yang_type)

    return query



from setuptools import setup


setup(
    name='alpakka',
    description="Cutest YANG Eater alive",

    author="ADVA Optical Networking",
    maintainer="Thomas Szyrkowiec",
    maintainer_email="tszyrkowiec@advaoptical.com",

    license="Apache License 2.0",

    setup_requires=open('requirements.setup.txt'),
    install_requires=['pyang', 'path.py', 'ipython'],

    use_scm_version={'local_scheme': lambda _: ''},

    packages=[
        'alpakka',
        'alpakka.pyang_plugins',
        'alpakka.wrapper',
        'alpakka.wools',
    ],
    entry_points={
        'console_scripts': ['alpakka=alpakka:run'],
    },

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Telecommunications Industry',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development',
        'Topic :: Software Development :: Code Generators',
        'Topic :: Utilities',
    ],
    keywords=[
        'alpakka', 'yang', 'pyang', 'wool', 'wools',
    ],
)



from alpakka.wrapper.nodewrapper import NodeWrapper
from alpakka.wrapper.nodewrapper import Listonder
from collections import OrderedDict
from alpakka.logger import LOGGER
import alpakka

WOOLS = alpakka.WOOLS


class Grouponder(NodeWrapper):
    """
    Base class for node wrappers that group variables.
    """

    def __init__(self, statement, parent):
        """
        Constructor for a grouponder which extracts all statements which are
        childes of the current statement

        :param children: all wrapped statements which are children of this
                         statement as list
        :param uses: if the current statement has a uses substmt a Grouping
                     object created and stored in the uses var
        :param is_augmented: identifier to indicate is the current statement
                             augmented by an other module or not
        """
        super().__init__(statement, parent)
        self.uses = OrderedDict()
        self.children = OrderedDict()

        # the following line separates stmts which are imported with 'uses'
        # from normal integrated stmts
        children_list = set(getattr(statement, 'i_children', ()))

        # wrap all children of the node
        for child in children_list:
            child_wrapper = self.WOOL.get(child.keyword)
            if child_wrapper:
                self.children[child.arg] = child_wrapper(child, parent=self)
            else:
                LOGGER.info("No wrapper for yang type: %s (%s)" %
                            (child.keyword, child.arg))

        # find all stmts which are imported with a 'uses' substmt and wrap the
        # Grouping object related to ths uses
        uses_list = set(statement.search('uses'))
        for stmt in uses_list:
            # check is the used grouping already wrapped
            # if so, the wrapped grouping is linked in the 'uses' variable
            # if not a the grouping statement is wrapped
            key = stmt.i_grouping.parent.arg + '/' + stmt.i_grouping.arg
            group = self.top().all_nodes.get('grouping', {}).get(key)
            if group:
                self.uses[group.yang_name()] = group
            else:
                self.uses[stmt.i_grouping.arg] = \
                    self.WOOL['grouping'](stmt.i_grouping, parent=self.top())

        # special handling for augment imports
        # collect all keys of augments

        augmented_keys = []

        for children in set(getattr(statement, 'i_children', ())):
            if hasattr(children, 'i_augment'):
                self.is_augmented = True
                augment_stmt = children.i_augment
                if augment_stmt.parent.arg not in augmented_keys:
                    augmented_keys.append(augment_stmt.parent.arg)
                    # Handling for Groupings which are imported with
                    # a augment statement
                    for grp in set(augment_stmt.search('uses')):
                        key = grp.parent.arg[1:] + '/' + \
                              grp.arg
                        group = self.top().all_nodes.get('grouping', {}). \
                            get(key)
                        if group:
                            self.uses[group.yang_name()] = group
                        else:
                            self.uses[grp.arg] = \
                                self.WOOL['grouping'](grp,
                                                      parent=self.top())

    def __getitem__(self, key):
        """
        Methode, that allows to access of the child elements directly as list
        of the top element
        :param key: name of the child elements which should be accessed
        :return: child object
        """

        return self.children[key]

    def __getattr__(self, name):
        """
        Methode, that allows to access a child element directly due the
        <parent>.<child> syntax

        :param name: name of the child element
        :return: child element object
        """
        if name == 'children':
            raise AttributeError
        key = name.replace('_', '-')
        try:
            return self.children[key]

        except KeyError:
            raise AttributeError("{!r} has no attribute {!r} or child {!r}"
                                 .format(self, name, key))

    def __dir__(self):
        """
        Mathode, that list all child elements as options of the auto completion

        :return: list of elements that are in the children list
        """
        return (*super().__dir__(),
                *(key.replace('-', '_') for key in
                  getattr(self, 'children', dict())))

    def all_children(self):
        """
        Collects a list of all child stmts of the current stmt regardless of
        the kind of implementation (local or import).
        :return: list of stmts
        """
        result = OrderedDict(self.children)
        result.update(self.uses)
        return result


class Module(Grouponder, yang='module'):
    """
    Wrapper class for a module statement.
    """

    def __init__(self, statement, parent=None):
        """
        Constructor for Class objects for module statements

        :param modules: Dictionary of classes of yang statements, where each
        class is a dictionary of all wrapped statements of the current module
        :param derived_types: Dictionary of all type defs which are used inside
        this module
        """
        self.all_nodes = {}
        self.derived_types = OrderedDict()
        super().__init__(statement, parent)
        # if self.statement.search_one('augment'):
        #     container = self.statement.search_one('augment').i_target_node
        #     self.WOOL.get(container.keyword)(container, self)


class Container(Grouponder, yang='container'):
    """
    Wrapper class for container statements.
    """

    def __init__(self, statement, parent):
        super().__init__(statement, parent)


class Grouping(Grouponder, yang='grouping'):
    """
    Wrapper class for grouping statements.
    """

    def __init__(self, statement, parent):
        super().__init__(statement, parent)


class List(Grouponder, Listonder, yang='list'):
    """
    Wrapper class for list statements.

    :param keys: the values originally stored in the key substmt of an yang
    list stmt
    """

    def __init__(self, statement, parent):
        super().__init__(statement, parent)
        self.keys = set()
        key_stmt = statement.search_one('key')
        if key_stmt is not None and key_stmt.arg is not None:
            # the key statement contains a space separated list of keys
            for key in key_stmt.arg.split():
                self.keys.add(key)


class Choice(Grouponder, yang='choice'):
    """
    Wrapper class for choice statements

    :param cases: Dictionary of all possible cases for the choice stmt
    """

    def __init__(self, statement, parent):
        super().__init__(statement, parent)
        self.cases = OrderedDict()
        for case in self.statement.search('case'):
            self.cases[case.arg] = self.WOOL['case'](case, self)


class Case(Grouponder, yang='case'):
    """
    Wrapper class for case statements
    """

    def __init__(self, statement, parent):
        super().__init__(statement, parent)


class RPC(Grouponder, yang='rpc'):
    """
    Parser for RPC nodes.

    :param input: stores the wrapped input statement
    :param output: stores the wrapped output statement
    """

    def __init__(self, statement, parent):
        super().__init__(statement, parent)
        for item in self.statement.i_children:
            if item.keyword == 'input':
                self.input = self.WOOL['input'](item, self)
            elif item.keyword == 'output':
                self.output = self.WOOL['output'](item, self)


class Input(Grouponder, yang='input'):
    """
    Parser for input nodes.
    """

    def __init__(self, statement, parent):
        super().__init__(statement, parent)


class Output(Grouponder, yang='output'):
    """
    Parser for output nodes.
    """

    def __init__(self, statement, parent):
        super().__init__(statement, parent)



from alpakka.wrapper.nodewrapper import NodeWrapper
from alpakka.wrapper.nodewrapper import Listonder
from collections import OrderedDict
import pyang.types


class Typonder(NodeWrapper):
    """
    Base class for node wrappers that have a type property.
    """

    def __init__(self, statement, parent):
        """
        Constructor for a typonder which extracts the data-type.

        :param data_type:
            string containing the argument of the type attribute of the yang
            statement
        :param is_build_in_type:
            boolean indicating is the data_type a yang base type or a typedef
            type
        :param enumeration:
            attribute containing the enumeration object if the data_type is
            enumeration
        :param union:
            attribute containing the union object if the data_type is union
        """
        super().__init__(statement, parent)
        type_stmt = statement.search_one('type')
        if type_stmt:
            data_types = self.WOOL.data_type_mappings
            # processing if the yang type is a base type
            if pyang.types.is_base_type(type_stmt.arg):
                self.data_type = data_types[type_stmt.arg]
                self.is_build_in_type = True
            # processing if the yang type is typedef
            else:
                self.data_type = type_stmt.arg
                # check is the typedef already generated
                if self.data_type not in self.top().derived_types.keys():
                    self.top().derived_types[self.data_type] = (
                        self.WOOL['typedef'](type_stmt.i_typedef,
                                             parent=self.top()))
                self.is_build_in_type = False

            # special processing if the data_type is enumeration
            if self.data_type == 'enumeration':
                self.enumeration = Enumeration(type_stmt, parent=self)
            # special processing if the data_type is an union
            elif self.data_type == 'union':
                self.union = Union(type_stmt, parent=self)
            elif self.data_type == 'leafref':
                if hasattr(type_stmt.i_type_spec, 'i_target_node'):
                    self.reference = self.WOOL['leaf'](
                        type_stmt.i_type_spec.i_target_node, self)
                self.path = next(
                    (i.substmts[0].arg for i in statement.substmts if
                     i.keyword == 'type'), None)

    def default_value(self):
        """
        Methode that returns the default value of a node if present

        :return: default value as string
        """
        for item in self.statement.search('default'):
            return item.arg

    def is_mandatory(self):
        """
        Methodes that returns the True if the node was selected to be
        mandatory and False if it is not mandatory
        :return:
        """
        for item in self.statement.search('mandatory'):
            return item.arg == 'true'

        return False


class Leaf(Typonder, yang='leaf'):
    """
    Wrapper class for type definition statements.

    :param path: used for leafrefs to store the reference path
    """

    def __init__(self, statement, parent):
        super().__init__(statement, parent)


class Enum(Typonder, yang='enum'):
    """
    Wrapper class for enum values.
    """

    def __init__(self, statement, parent):
        super().__init__(statement, parent)


class Enumeration(Typonder, yang='enumeration'):
    """
    Wrapper class for enumeration statements.

    :param enums: Dictionary to store all wrapped enum objects
    """

    def __init__(self, statement, parent):
        super().__init__(statement, parent)
        self.enums = OrderedDict()
        # loop through substatements and extract the enum values
        for stmt in statement.search('enum'):
            self.enums[stmt.arg] = self.WOOL['enum'](stmt, self)


class Union(Typonder, yang='union'):
    """
    Wrapper class for union statements

    :param types: Dictionary to store all types which are part of the union,
    stored as string or wrapped statement object
    """

    def __init__(self, statement, parent):
        super().__init__(statement, parent)
        # list of types that belong to the union
        self.types = OrderedDict()
        for stmt in statement.search('type'):
            if pyang.types.is_base_type(stmt.arg):
                wool_data_type = self.WOOL.data_type_mappings[stmt.arg]
                self.types[wool_data_type] = wool_data_type
            elif stmt.arg in self.top().derived_types.keys():
                key = stmt.arg
                self.types[stmt.arg] = self.top().derived_types.get(
                    key) or stmt.arg
            else:
                self.types[stmt.arg] = self.WOOL['typedef'](stmt,
                                                            parent=self.top())


class TypeDef(Typonder, yang='typedef'):
    """
    Wrapper class for type definition statements.
    """

    def __init__(self, statement, parent):
        super().__init__(statement, parent)


class LeafList(Typonder, Listonder, yang='leaf-list'):
    """
    Wrapper class for leaf list statements.

    :param path: used for leafrefs to store the reference path
    """

    def __init__(self, statement, parent):
        super().__init__(statement, parent)
